@@ -1,38 +1,39 @@
+#  Licensed under the Apache License, Version 2.0 (the "License"); you may
+#  not use this file except in compliance with the License. You may obtain
+#  a copy of the License at
+#
+#       http://www.apache.org/licenses/LICENSE-2.0
+#
+#  Unless required by applicable law or agreed to in writing, software
+#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
+#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
+#  License for the specific language governing permissions and limitations
+#  under the License.
+
 """Application base class.
 """
 
-import argparse
-import codecs
 import inspect
 import locale
 import logging
 import logging.handlers
 import os
 import sys
-import operator
 
-from .complete import CompleteCommand
-from .help import HelpAction, HelpCommand
-from .utils import damerau_levenshtein, COST
+import cmd2
 
-# Make sure the cliff library has a logging handler
-# in case the app developer doesn't set up logging.
-# For py26 compat, create a NullHandler
+from cliff import _argparse
+from . import complete
+from . import help
+from . import utils
 
-if hasattr(logging, 'NullHandler'):
-    NullHandler = logging.NullHandler
-else:
-    class NullHandler(logging.Handler):
-        def handle(self, record):
-            pass
 
-        def emit(self, record):
-            pass
+logging.getLogger('cliff').addHandler(logging.NullHandler())
 
-        def createLock(self):
-            self.lock = None
 
-logging.getLogger('cliff').addHandler(NullHandler())
+# Exit code for exiting due to a signal is 128 + the signal number
+_SIGINT_EXIT = 130
+_SIGPIPE_EXIT = 141
 
 
 class App(object):
@@ -53,9 +54,14 @@ class App(object):
     :param interactive_app_factory: callable to create an
                                     interactive application
     :paramtype interactive_app_factory: cliff.interactive.InteractiveApp
+    :param deferred_help: True - Allow subcommands to accept --help with
+                          allowing to defer help print after initialize_app
+    :paramtype deferred_help: bool
     """
 
     NAME = os.path.splitext(os.path.basename(sys.argv[0]))[0]
+    if NAME == '__main__':
+        NAME = os.path.split(os.path.dirname(sys.argv[0]))[-1]
     LOG = logging.getLogger(NAME)
 
     CONSOLE_MESSAGE_FORMAT = '%(message)s'
@@ -71,8 +77,8 @@ class App(object):
         """Initialize the application.
         """
         self.command_manager = command_manager
-        self.command_manager.add_command('help', HelpCommand)
-        self.command_manager.add_command('complete', CompleteCommand)
+        self.command_manager.add_command('help', help.HelpCommand)
+        self.command_manager.add_command('complete', complete.CompleteCommand)
         self._set_streams(stdin, stdout, stderr)
         self.interactive_app_factory = interactive_app_factory
         self.deferred_help = deferred_help
@@ -85,24 +91,36 @@ class App(object):
             locale.setlocale(locale.LC_ALL, '')
         except locale.Error:
             pass
-        if sys.version_info[:2] == (2, 6):
-            # Configure the input and output streams. If a stream is
-            # provided, it must be configured correctly by the
-            # caller. If not, make sure the versions of the standard
-            # streams used by default are wrapped with encodings. This
-            # works around a problem with Python 2.6 fixed in 2.7 and
-            # later (http://hg.python.org/cpython/rev/e60ef17561dc/).
-            lang, encoding = locale.getdefaultlocale()
-            encoding = (getattr(sys.stdout, 'encoding', None) or
-                        encoding or
-                        self.DEFAULT_OUTPUT_ENCODING)
-            self.stdin = stdin or codecs.getreader(encoding)(sys.stdin)
-            self.stdout = stdout or codecs.getwriter(encoding)(sys.stdout)
-            self.stderr = stderr or codecs.getwriter(encoding)(sys.stderr)
-        else:
-            self.stdin = stdin or sys.stdin
-            self.stdout = stdout or sys.stdout
-            self.stderr = stderr or sys.stderr
+
+        # Unicode must be encoded/decoded for text I/O streams, the
+        # correct encoding for the stream must be selected and it must
+        # be capable of handling the set of characters in the stream
+        # or Python will raise a codec error. The correct codec is
+        # selected based on the locale. Python2 uses the locales
+        # encoding but only when the I/O stream is attached to a
+        # terminal (TTY) otherwise it uses the default ASCII
+        # encoding. The effect is internationalized text written to
+        # the terminal works as expected but if command line output is
+        # redirected (file or pipe) the ASCII codec is used and the
+        # program aborts with a codec error.
+        #
+        # The default I/O streams stdin, stdout and stderr can be
+        # wrapped in a codec based on the locale thus assuring the
+        # users desired encoding is always used no matter the I/O
+        # destination. Python3 does this by default.
+        #
+        # If the caller supplies an I/O stream we use it unmodified on
+        # the assumption the caller has taken all responsibility for
+        # the stream.  But with Python2 if the caller allows us to
+        # default the I/O streams to sys.stdin, sys.stdout and
+        # sys.stderr we apply the locales encoding just as Python3
+        # would do. We also check to make sure the main Python program
+        # has not already already wrapped sys.stdin, sys.stdout and
+        # sys.stderr as this is a common recommendation.
+
+        self.stdin = stdin or sys.stdin
+        self.stdout = stdout or sys.stdout
+        self.stderr = stderr or sys.stderr
 
     def build_option_parser(self, description, version,
                             argparse_kwargs=None):
@@ -120,7 +138,7 @@ class App(object):
         :paramtype extra_kwargs: dict
         """
         argparse_kwargs = argparse_kwargs or {}
-        parser = argparse.ArgumentParser(
+        parser = _argparse.ArgumentParser(
             description=description,
             add_help=False,
             **argparse_kwargs
@@ -128,48 +146,49 @@ class App(object):
         parser.add_argument(
             '--version',
             action='version',
-            version='%(prog)s {0}'.format(version),
+            version='{0} {1}'.format(App.NAME, version),
         )
-        parser.add_argument(
+        verbose_group = parser.add_mutually_exclusive_group()
+        verbose_group.add_argument(
             '-v', '--verbose',
             action='count',
             dest='verbose_level',
             default=self.DEFAULT_VERBOSE_LEVEL,
             help='Increase verbosity of output. Can be repeated.',
         )
+        verbose_group.add_argument(
+            '-q', '--quiet',
+            action='store_const',
+            dest='verbose_level',
+            const=0,
+            help='Suppress output except warnings and errors.',
+        )
         parser.add_argument(
             '--log-file',
             action='store',
             default=None,
             help='Specify a file to log output. Disabled by default.',
         )
-        parser.add_argument(
-            '-q', '--quiet',
-            action='store_const',
-            dest='verbose_level',
-            const=0,
-            help='suppress output except warnings and errors',
-        )
         if self.deferred_help:
             parser.add_argument(
                 '-h', '--help',
                 dest='deferred_help',
                 action='store_true',
-                help="show this help message and exit",
+                help="Show help message and exit.",
             )
         else:
             parser.add_argument(
                 '-h', '--help',
-                action=HelpAction,
+                action=help.HelpAction,
                 nargs=0,
                 default=self,  # tricky
-                help="show this help message and exit",
+                help="Show help message and exit.",
             )
         parser.add_argument(
             '--debug',
             default=False,
             action='store_true',
-            help='show tracebacks on errors',
+            help='Show tracebacks on errors.',
         )
         return parser
 
@@ -210,7 +229,7 @@ class App(object):
            explicitly this method in initialize_app.
         """
         if self.deferred_help and self.options.deferred_help:
-            action = HelpAction(None, None, default=self)
+            action = help.HelpAction(None, None, default=self)
             action(self.parser, self.options, None, None)
 
     def run(self, argv):
@@ -219,6 +238,7 @@ class App(object):
         :param argv: input arguments and options
         :paramtype argv: list of str
         """
+
         try:
             self.options, remainder = self.parser.parse_known_args(argv)
             self.configure_logging()
@@ -237,6 +257,8 @@ class App(object):
                 remainder.insert(0, "help")
             self.initialize_app(remainder)
             self.print_help_if_requested()
+        except BrokenPipeError:
+            return _SIGPIPE_EXIT
         except Exception as err:
             if hasattr(self, 'options'):
                 debug = self.options.debug
@@ -248,11 +270,18 @@ class App(object):
             else:
                 self.LOG.error(err)
             return 1
+        except KeyboardInterrupt:
+            return _SIGINT_EXIT
         result = 1
         if self.interactive_mode:
             result = self.interact()
         else:
-            result = self.run_subcommand(remainder)
+            try:
+                result = self.run_subcommand(remainder)
+            except BrokenPipeError:
+                return _SIGPIPE_EXIT
+            except KeyboardInterrupt:
+                return _SIGINT_EXIT
         return result
 
     # FIXME(dhellmann): Consider moving these command handling methods
@@ -298,8 +327,7 @@ class App(object):
                                                         self.stdin,
                                                         self.stdout,
                                                         )
-        self.interpreter.cmdloop()
-        return 0
+        return self.interpreter.cmdloop()
 
     def get_fuzzy_matches(self, cmd):
         """return fuzzy matches of unknown command
@@ -314,21 +342,23 @@ class App(object):
             prefix = candidate.split(sep)[0]
             # Give prefix match a very good score
             if candidate.startswith(cmd):
-                dist.append((candidate, 0))
+                dist.append((0, candidate))
                 continue
             # Levenshtein distance
-            dist.append((candidate, damerau_levenshtein(cmd, prefix, COST)+1))
-        dist = sorted(dist, key=operator.itemgetter(1, 0))
+            dist.append((utils.damerau_levenshtein(cmd, prefix, utils.COST)+1,
+                         candidate))
+
         matches = []
-        i = 0
-        # Find the best similarity
-        while (not dist[i][1]):
-            matches.append(dist[i][0])
-            i += 1
-        best_similarity = dist[i][1]
-        while (dist[i][1] == best_similarity):
-            matches.append(dist[i][0])
-            i += 1
+        match_distance = 0
+        for distance, candidate in sorted(dist):
+            if distance > match_distance:
+                if match_distance:
+                    # we copied all items with minimum distance, we are done
+                    break
+                # we copied all items with distance=0,
+                # now we match all candidates at the minimum distance
+                match_distance = distance
+            matches.append(candidate)
 
         return matches
 
@@ -345,7 +375,7 @@ class App(object):
                     article = 'an'
                 self.stdout.write('%s: \'%s\' is not %s %s command. '
                                   'See \'%s --help\'.\n'
-                                  % (self.NAME, the_cmd, article,
+                                  % (self.NAME, ' '.join(argv), article,
                                       self.NAME, self.NAME))
                 self.stdout.write('Did you mean one of these?\n')
                 for match in fuzzy_matches:
@@ -358,11 +388,11 @@ class App(object):
             return 2
         cmd_factory, cmd_name, sub_argv = subcommand
         kwargs = {}
-        if 'cmd_name' in inspect.getargspec(cmd_factory.__init__).args:
+        if 'cmd_name' in inspect.getfullargspec(cmd_factory.__init__).args:
             kwargs['cmd_name'] = cmd_name
         cmd = cmd_factory(self, self.options, **kwargs)
-        err = None
         result = 1
+        err = None
         try:
             self.prepare_to_run_command(cmd)
             full_name = (cmd_name
@@ -370,13 +400,28 @@ class App(object):
                          else ' '.join([self.NAME, cmd_name])
                          )
             cmd_parser = cmd.get_parser(full_name)
-            parsed_args = cmd_parser.parse_args(sub_argv)
+            try:
+                parsed_args = cmd_parser.parse_args(sub_argv)
+            except SystemExit as ex:
+                raise cmd2.exceptions.Cmd2ArgparseError from ex
             result = cmd.run(parsed_args)
-        except Exception as err:
+        except BrokenPipeError as err1:
+            result = _SIGPIPE_EXIT
+            err = err1
+            raise
+        except help.HelpExit:
+            result = 0
+        except Exception as err1:
+            err = err1
             if self.options.debug:
                 self.LOG.exception(err)
             else:
                 self.LOG.error(err)
+        except KeyboardInterrupt as err1:
+            result = _SIGINT_EXIT
+            err = err1
+            raise
+        finally:
             try:
                 self.clean_up(cmd, result, err)
             except Exception as err2:
@@ -384,14 +429,5 @@ class App(object):
                     self.LOG.exception(err2)
                 else:
                     self.LOG.error('Could not clean up: %s', err2)
-            if self.options.debug:
-                raise
-        else:
-            try:
-                self.clean_up(cmd, result, None)
-            except Exception as err3:
-                if self.options.debug:
-                    self.LOG.exception(err3)
-                else:
-                    self.LOG.error('Could not clean up: %s', err3)
+            del err
         return result
