@@ -1,12 +1,13 @@
 #!/usr/bin/env python
 
+from setuptools import find_packages
+from setuptools import setup
+
 PROJECT = 'cliffdemo'
 
 # Change docs/sphinx/conf.py too!
 VERSION = '0.1'
 
-from setuptools import setup, find_packages
-
 try:
     long_description = open('README.rst', 'rt').read()
 except IOError:
@@ -25,16 +26,15 @@ setup(
     url='https://github.com/openstack/cliff',
     download_url='https://github.com/openstack/cliff/tarball/master',
 
-    classifiers=['Development Status :: 3 - Alpha',
-                 'License :: OSI Approved :: Apache Software License',
-                 'Programming Language :: Python',
-                 'Programming Language :: Python :: 2',
-                 'Programming Language :: Python :: 2.7',
-                 'Programming Language :: Python :: 3',
-                 'Programming Language :: Python :: 3.2',
-                 'Intended Audience :: Developers',
-                 'Environment :: Console',
-                 ],
+    classifiers=[
+        'Development Status :: 3 - Alpha',
+        'License :: OSI Approved :: Apache Software License',
+        'Programming Language :: Python',
+        'Programming Language :: Python :: 3',
+        'Programming Language :: Python :: 3 :: Only',
+        'Intended Audience :: Developers',
+        'Environment :: Console',
+    ],
 
     platforms=['Any'],
 
@@ -60,6 +60,10 @@ setup(
             'file = cliffdemo.show:File',
             'show file = cliffdemo.show:File',
             'unicode = cliffdemo.encoding:Encoding',
+            'hooked = cliffdemo.hook:Hooked',
+        ],
+        'cliff.demo.hooked': [
+            'sample-hook = cliffdemo.hook:Hook',
         ],
     },
 
